/*
** Module   : process_win.c
** Abstract : implementation for process creation/waiting logic in p2j.dll native library which
**            is called via JNI
**            This is used by process.c
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130329 Created initial version to implement Windows specifics for process launching
**                  operations.
** 002 EVL 20130829 Modifications allowing starting processes from Windows console driver.
*/

#include "process_launch.h" 
#include <sys/types.h>
#include <windows.h>
#include <fcntl.h>

/**
 * initializeFileHandle - initialize the passed file handle object with the pty fh
 *
 * @param   env         pointer to an object providing access to JNI environment

 * @param   descriptor  object representing the FileDescriptor Java instance 
 *
 * @param   fh          the pty file handle
 *
 */ 
void initializeFileHandle(JNIEnv* env, jobject descriptor, long fh)
{
   jclass fdcls = (*env)->GetObjectClass(env, descriptor);
   jfieldID jhandle = (*env)->GetFieldID(env, fdcls, "handle", "J");

   (*env)->SetLongField(env, descriptor, jhandle, fh);
}

/**
 * clearScreen - clear the entire screen. This is replacement for NCURSES functionality.
 *
 * @param   hConsole    handle to console to clean up the screen.
 *
 */ 
void clearScreen(HANDLE hConsole)
{
   COORD coordsIni = {0, 0};        // Initial cursor position 
   DWORD charsWritten;              // The nimber of chars written
   CONSOLE_SCREEN_BUFFER_INFO csbi; // Screen buffer information
   DWORD conSize;                   // The size of console

   // Get the number of character cells in the current buffer. 
   if (!GetConsoleScreenBufferInfo(hConsole, &csbi))
   {
      return;
   }

   conSize = csbi.dwSize.X * csbi.dwSize.Y;

   // Fill the entire screen with blanks.
   if (!FillConsoleOutputCharacter(hConsole,       // Handle to console screen buffer 
                                   (TCHAR) ' ',    // Character to write to the buffer
                                   conSize,        // Number of cells to write 
                                   coordsIni,      // Coordinates of first cell 
                                   &charsWritten)) // Receive number of characters written
   {
      return;
   }

   // Get the current text attribute.
   if (!GetConsoleScreenBufferInfo(hConsole, &csbi))
   {
      return;
   }
   
   // Set the buffer's attributes accordingly.
   if (!FillConsoleOutputAttribute(hConsole,         // Handle to console screen buffer 
                                   csbi.wAttributes, // Character attributes to use
                                   conSize,          // Number of cells to set attribute 
                                   coordsIni,        // Coordinates of first cell 
                                   &charsWritten))   // Receive number of characters written
   {
      return;
   }

   // Put the cursor at its home coordinates.
   SetConsoleCursorPosition(hConsole, coordsIni);
}

/**
 * init - Windows version of set up signal handling to ignore job control signals
 *        (reflecteds in the child process as well)
 *
 * @param   env   pointer to an object providing access to JNI environment
 *
 * @param   cls   reference to the Java class who called this function
 *                (from static context)
 *
 */ 
JNIEXPORT void JNICALL init(JNIEnv* env, jclass class)
{
   // The function sigemptyset() is not supported in Windows due to Windows signal paradigm does
   // not have a mask and what is more implortant we do not need to prepare structure used in
   // this function under Linux in Windows version if the init call.
   
   // Setting up signal handlers
   signal(SIGINT,  SIG_IGN);
   // The signals SIGTSTP, SIGTTIN, SIGTTOU are missing in Windows meaning ignore by default
   // This means in turn we do not need special calls to make them ignored.
}

/**
 * pseudoTerminalLaunch - spawns a child process redirecting its standard input, output and
 *                        error to particular file descriptors (FileDescriptor Java instances). 
 *                        Windows version.
 *
 * @param   env       pointer to a structure that provides access to JNI
 *                    environment
 *
 * @param   cls       reference to the Java class who called this function
 *                    (from static context)
 *
 * @param   cmdline   the command line used for launching the child process  
 *
 * @param   in        FileDescriptor Java object for stdin
 *
 * @param   out       FileDescriptor Java object for stdout
 *
 * @param   err       FileDescriptor Java object for stderr
 *
 *
 * @param   silent  boolean flag controlling interactive mode
 *
 * @return  > 1  PID of spawned child process if successful
 *          < 0  on error
 */ 

JNIEXPORT jint JNICALL pseudoTerminalLaunch(JNIEnv* env, jclass cls, jobjectArray cmdline,
                                            jobject in, jobject out, jobject err, jboolean silent)
   
{
   jint pid = 0;
   
   // make sure we have valid argument references
   if (cmdline == NULL)
   {
      return -1;
   }
   
   // make sure we have a tty as stdin
   if (!isatty(STDIN_FILENO))
   {
      return -2;
   }
   
   // access redirection flags and create pipes
   SECURITY_ATTRIBUTES secAttr;
   secAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
   secAttr.bInheritHandle = TRUE;
   secAttr.lpSecurityDescriptor = NULL;
   
   HANDLE pipe_i[2];                // pipe for stdin
   if (in != NULL)
   {
      if (!CreatePipe(&pipe_i[0], &pipe_i[1], &secAttr, 0))
      {
         return -3;
      }

      // Prepare input pipe for redirection, write end pipe inheritance for child process
      // standard input stream should be turned off
      if (!SetHandleInformation(pipe_i[1], HANDLE_FLAG_INHERIT, 0))
      {
         // Close the created pipe before abnormal end
         CloseHandle(pipe_i[0]);
         CloseHandle(pipe_i[1]);
         exit(-1);
      }
   }
   
   HANDLE pipe_o[2];                // pipe for stdout
   if (out != NULL)
   {
      if (!CreatePipe(&pipe_o[0], &pipe_o[1], &secAttr, 0))
      {
         if (in != NULL)
         {
            CloseHandle(pipe_i[0]);
            CloseHandle(pipe_i[1]);
         }
         return -4;
      }
      
      // Prepare output pipe for redirection, read end pipe inheritance for child process
      // standard input stream should be turned off
      if (!SetHandleInformation(pipe_o[0], HANDLE_FLAG_INHERIT, 0))
      {
         // Close the created pipes before abnormal end
         if (in != NULL)
         {
            CloseHandle(pipe_i[0]);
            CloseHandle(pipe_i[1]);
         }
         CloseHandle(pipe_o[0]);
         CloseHandle(pipe_o[1]);
         // terminating
         exit(-1);
      }
   }
   
   HANDLE pipe_e[2];                // pipe for stderr
   if (err != NULL)
   {
      if (!CreatePipe(&pipe_e[0], &pipe_e[1], &secAttr, 0))
      {
         if (in != NULL)
         {
            CloseHandle(pipe_i[0]);
            CloseHandle(pipe_i[1]);
         }
         if (out != NULL)
         {
            CloseHandle(pipe_o[0]);
            CloseHandle(pipe_o[1]);
         }
         return -5;
      }

      // Prepare error pipe for redirection, read end pipe inheritance for child process
      // standard error stream should be turned off
      if (!SetHandleInformation(pipe_e[0], HANDLE_FLAG_INHERIT, 0))
      {
         // Close opened pipes before abnormal end 
         if (in != NULL)
         {
            CloseHandle(pipe_i[0]);
            CloseHandle(pipe_i[1]);
         }
         if (out != NULL)
         {
            CloseHandle(pipe_o[0]);
            CloseHandle(pipe_o[1]);
         }
         CloseHandle(pipe_e[0]);
         CloseHandle(pipe_e[1]);
         // terminating
         exit(-1);
      }
   }

   // Need to inherit some pipes
   BOOL bInheritHandles = (in != NULL) | (out != NULL) | (err != NULL);

   // Minimal argument preparations
   STARTUPINFO sinfo;
   ZeroMemory( &sinfo, sizeof(sinfo) );
   sinfo.cb = sizeof(sinfo);
   GetStartupInfo(&sinfo);
   // Substitute child's standard handles with our pipes
   // The file descriptors are not the same as handles
   if (bInheritHandles)
   {
      if (in != NULL)
      {
         sinfo.hStdInput = pipe_i[0];
      }
      else
      {
         sinfo.hStdInput = GetStdHandle(STD_INPUT_HANDLE);
      }
      if (out != NULL)
      {
         sinfo.hStdOutput = pipe_o[1];
      }
      else
      {
         sinfo.hStdOutput = GetStdHandle(STD_OUTPUT_HANDLE);
      }
      if (err != NULL)
      {
         sinfo.hStdError = pipe_e[1];
      }
      else
      {
         sinfo.hStdError = GetStdHandle(STD_ERROR_HANDLE);
      }
      sinfo.dwFlags |= STARTF_USESTDHANDLES;
      // The options below suppress the console show for redirected i/o opeations 
      sinfo.dwFlags |= STARTF_USESHOWWINDOW;
      sinfo.wShowWindow = SW_HIDE;
   }

   // Minimal argument preparations
   PROCESS_INFORMATION pinfo;
   ZeroMemory( &pinfo, sizeof(pinfo) );

   // must build the parameter list
   int len = (int) (*env)->GetArrayLength(env, cmdline);
   int i;
   int finalLength = 0;

   // Getting the total number of characters in final string
   for (i = 0; i < len; i++)
   {
      // Getting next character part from java array
      jobject next = (*env)->GetObjectArrayElement(env, cmdline, i);
      // Do not make a copy inside Java code for string, use original
      const char * tmp  = (char*) (*env)->GetStringUTFChars(env, (jstring) next, NULL);
      if (tmp == NULL)
      {
         // Close opened pipes before exiting 
         if (in != NULL)
         {
            CloseHandle(pipe_i[0]);
            CloseHandle(pipe_i[1]);
         }
         if (out != NULL)
         {
            CloseHandle(pipe_o[0]);
            CloseHandle(pipe_o[1]);
         }
         if (err != NULL)
         {
            CloseHandle(pipe_e[0]);
            CloseHandle(pipe_e[1]);
         }
         // Out of memory! Terminating
         exit(-1);
      }

      // For second and more part we need to insert space
      if (i > 0)
      {
         finalLength += 1;
      }
      finalLength += strlen(tmp);

      // Free java memory
      (*env)->ReleaseStringUTFChars(env, (jstring) next, tmp);
   }
   // One more character for trailing null terminatror
   finalLength += 1;

   // Allocate memory from heap to command line and one null terminator   
   char * cmds = malloc(sizeof(char) * finalLength);
   // Initialize the command line string with null terminated empty
   strcpy(cmds, "");
   // copy the strings into our own array
   for (i = 0; i < len; i++)
   {
      // Getting next character part from java array
      jobject next = (*env)->GetObjectArrayElement(env, cmdline, i); 
      // Do not make a copy inside Java code for string, use original
      const char * tmp  = (char*) (*env)->GetStringUTFChars(env, (jstring) next, NULL);
         
      if (tmp == NULL)
      {
         // Close opened pipes before exiting 
         if (in != NULL)
         {
            CloseHandle(pipe_i[0]);
            CloseHandle(pipe_i[1]);
         }
         if (out != NULL)
         {
            CloseHandle(pipe_o[0]);
            CloseHandle(pipe_o[1]);
         }
         if (err != NULL)
         {
            CloseHandle(pipe_e[0]);
            CloseHandle(pipe_e[1]);
         }
         // Free local heap allocation
         free(cmds);
         // Out of memory! Terminating
         exit(-1);
      }

      // For second and more part we need to insert space
      if (i > 0)
      {
         strcat(cmds, " ");
      }
      // Append next command line part
      strcat(cmds, tmp);
      if (i == 1 && !bInheritHandles && !silent && cmds[strlen(cmds)-1] == 'c')
      {
         cmds[strlen(cmds)-1] = 'k';
      }
      
      // Free java memory
      (*env)->ReleaseStringUTFChars(env, (jstring) next, tmp);
   }

   // This is the title to display in newly created console
   sinfo.lpTitle = "cmd.exe";
   // create a child process
   if (!CreateProcess(NULL, cmds, NULL, NULL, bInheritHandles, CREATE_NEW_CONSOLE, NULL, NULL,
                      &sinfo, &pinfo))
   {
      // Process creation failed, close every opened pipes
      if (in != NULL)
      {
         CloseHandle(pipe_i[0]);
         CloseHandle(pipe_i[1]);
      }
      if (out != NULL)
      {
         CloseHandle(pipe_o[0]);
         CloseHandle(pipe_o[1]);
      }
      if (err != NULL)
      {
         CloseHandle(pipe_e[0]);
         CloseHandle(pipe_e[1]);
      }
      // Free local heap allocation
      free(cmds);
      return (jint) -6;
   }
   else
   {
      // Return PID counterpart
      // Yes this is 64->32 bit truncation but for handles it is safe
      pid = (jint) HandleToLong(pinfo.hProcess);

      // parent process
      // initialize the file descriptors for the redirected handles
      if (in != NULL)
      {
         // don't need read end for child's stdin in parent
         CloseHandle(pipe_i[0]);
         
         // pass the write end back
         initializeFileHandle(env, in, HandleToLong(pipe_i[1]));
      }
      
      if (out != NULL)
      {
         // don't need write end for child's stdout in parent
         CloseHandle(pipe_o[1]);
         
         // pass the read end back
         initializeFileHandle(env, out, HandleToLong(pipe_o[0]));
      }
      
      if (err != NULL)
      {
         // don't need write end for child's stderr in parent
         CloseHandle(pipe_e[1]);
         
         // pass the read end back
         initializeFileHandle(env, err, HandleToLong(pipe_e[0]));
      }
   }
   // Free local heap allocation
   free(cmds);
   return pid;
}

/**
 * pseudoTerminalWait - The Windows version of waits for a child process to terminate. 
 *
 * @param   env     pointer to a structure that provides access to JNI environment
 *
 * @param   cls     reference to the Java class who called this function from
 *                  a static context
 *
 * @param   pid     PID of the process to be waited
 *
 * @param   silent  boolean flag to control interactive mode
 *
 */ 
JNIEXPORT void JNICALL pseudoTerminalWait(JNIEnv* env, jclass cls, jint pid, jboolean silent)
{
   HANDLE hProcess = LongToHandle(pid);

   // Wait for the process to finish with completion I/O operations
   WaitForSingleObjectEx(hProcess, INFINITE, TRUE);
   // Closing the process handle
   CloseHandle(hProcess);
   
   // Implementation for interactive terminals
   if (!silent)
   {
      // in interactive mode we're calling functions in NCURSES directly, 
      // as opposed to executing command line utilities for the same purpose
      // (NCURSES library is a requirement anyway so we want to have a more
      // efficient and more controlled use of that interface as well, while 
      // also avoiding the hard requirement that other utility programs be
      // installed in addition to P2J)

      // put "Press space bar to continue." message on the last line
      // Clear terminal screen
      clearScreen(GetStdHandle(STD_OUTPUT_HANDLE));
   }
}

